#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <time.h>

#ifdef VMS
  #include <ioctl.h>
#endif

#ifdef LINUX
  #include <fcntl.h>
#endif

#ifdef SUN
  #include <sys/filio.h>
#endif

#ifdef borland
  #include <io.h>
  #include <winsock.h>
#else
  #include <sys/socket.h>
  #include <sys/types.h>
  #include <netinet/in.h>
  #include <netdb.h>
  #include <unistd.h>
#endif

#define RECORD_SIZE 512
#define LF "\n"
#define CR "\r"


#ifdef borland
  #define close closesocket
#endif  


/*
 *   FUNCTION NAME
 *    srvconn
 *
 *   DESCRIPTION
 *    connect to the server.
 *
 *   ARGUMENTS
 *    host - host name
 *    port - port number
 *    blocking - true: connect blocks
 *    sock - socket value
 *
 *   RETURN VALUE
 *    Boolean error flag (0 = error)
 *
 *   GLOBALS, ENVIRONMENT VARIABLES, FILE
 * 
 *   REMARK 
 *    a new socket is only created if sock = 0
 *
 *   COMMENTS
 */

int srvconn (char *host, int port, unsigned long blocking, int *sock)
{
  struct sockaddr_in sin;
  struct hostent     *hp;
  struct in_addr     **adr_zgr;
  int                fsflags;
  int                error;
  unsigned long      mode;
  char               *s;


  /* Create socket */
  if ((*sock = socket (PF_INET, SOCK_STREAM, 0)) == -1) {
#ifdef borland
    error = WSAGetLastError();
    fprintf(stderr,"***Error creating socket!*** %d\n",error);
#else    
    perror("");
#endif      
    return 0;
  }

  /* Initialize and get connection */  
  memset((char *)&sin, '\0', sizeof(sin));
  sin.sin_family = AF_INET;

#ifdef debug
    fprintf(stdout,"Portnumber: %d\n",port);
#endif

  sin.sin_port = htons (port);
  hp = gethostbyname (host);
  if (hp == NULL) {
    fprintf(stderr,"***Error searching host name***\n");
    return 0;
  }

#ifdef borland
  memcpy ((char *) &sin.sin_addr, (char *) hp->h_addr, hp->h_length);
#else
  memcpy (&sin.sin_addr, hp->h_addr, sizeof(struct in_addr));
#endif

#ifdef debug
  fprintf(stdout,"Address: %s\n",inet_ntoa(sin.sin_addr)); 
#endif

#ifdef borland
  // set socket blocking/nonblocking
  if (blocking)
    mode = 0;
  else 
    mode = 1;  
  if (ioctlsocket(*sock,FIONBIO,&mode) < 0) {
    error = WSAGetLastError();
    fprintf(stderr,"***Error setting options*** %d\n",error);
    return 0;
  }

  if (connect (*sock, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
    error = WSAGetLastError();
    if ((error != WSAEWOULDBLOCK) && (error != WSAEALREADY) 
    && (error != WSAEINPROGRESS) && (error != WSAEISCONN)) {
      fprintf(stderr,"***Error Connecting Socket*** %d\n",error);
      close(*sock);
      return 0;
    }
  }
#else

#ifdef LINUX
  /* set socket options */
  if (fsflags = fcntl(*sock,F_GETFL,0) < 0) {
    perror("***Error getting options***");
    return 0;
  };

  if (blocking)
    fsflags &= ~O_NONBLOCK;
  else
    fsflags |= O_NONBLOCK;
		
  if (fcntl(*sock,F_SETFL,fsflags) < 0) {
    perror("***Error setting options***");
    return 0;
  }
#else
  if (blocking)
    mode = 0;
  else 
    mode = 1;  
  if (ioctl(*sock,FIONBIO,&mode) < 0) {
    perror("***Error setting options***");
    return 0;
  }
#endif

  if (connect (*sock, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
    if ((errno != EINPROGRESS) && (errno != EALREADY) 
    && (errno != EISCONN)) {
      perror("***");
      close(*sock);
      sock = 0;
      return 0;
    }
  }
#endif

  return 1;
}


/*
 *   FUNCTION NAME
 *    say
 *
 *   DESCRIPTION
 *    write a message or data on socket.
 *
 *   ARGUMENTS
 *    senddata - The message to write on socket
 *    timeout - Timeout in s for TCP/IP
 *    blocking -  0: socket blocking
 *               >0: socket non blocking
 *    socket: TCP/IP socket
 *
 *   RETURN VALUE
 *    -1 = error; 
 *     0 = not all bytes sent
 *     1 = ok;
 *
 *   GLOBALS, ENVIRONMENT VARIABLES, FILE
 *
 *   REMARK:
 *     
 *
 */

int say(int socket, double timeout, unsigned long blocking, char* senddata) {

  int             bytessend,nchar,bufferlen;
  double          sec;
  char*           buffer;
  double          eltime;
  int             fsflags;
  fd_set          writefds,exceptfds;
  time_t          start,end;
  struct timeval  seltimeout;
  int             error;

#ifdef debug
  printf("send: %s\n",senddata);
#endif

  /* add LF */
  if (strcmp(senddata+(strlen(senddata)-1),LF)!=0) {
    strcat(senddata,LF);
  }
  buffer = malloc(strlen(senddata)+1);
  strcpy(buffer,senddata);
  bufferlen = strlen(buffer);

  bytessend = 0;
  
  /* set socket options */ 
#ifdef borland
  if (ioctlsocket(socket,FIONBIO,&blocking) < 0) {
    fprintf(stderr,"***Error setting options***\n");
  }
#else
#ifdef LINUX
  /* set socket options */
  if (fsflags = fcntl(socket,F_GETFL,0) < 0) {
    perror("***Error getting options***");
    return 0;
  };

  if (blocking)
    fsflags &= ~O_NONBLOCK;
  else
    fsflags |= O_NONBLOCK;
		
  if (fcntl(socket,F_SETFL,fsflags) < 0) {
    perror("***Error setting options***");
    return 0;
  }
#else 
  if (ioctl(socket,FIONBIO,&blocking) < 0) {
    perror("***");
  }
#endif
#endif

  /* initialize writefds,exceptfds */
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);

  /* check timeout */
  if (timeout) { 
    time(&start);
    eltime = 0;
  } else {  
    eltime = -1;    
    seltimeout.tv_usec=0;
    seltimeout.tv_sec=0;
  }

  while (eltime < timeout) {

    if (timeout) {
      seltimeout.tv_usec=(int) floor(modf(timeout - eltime,&sec)*1000000);
      seltimeout.tv_sec=(int) sec;
    }

    FD_SET(socket,&writefds);

#ifdef borland
    FD_SET(socket,&exceptfds);
    if (select(socket+1,NULL,&writefds,&exceptfds,&seltimeout) < 0) {
      error = WSAGetLastError();
      if (error != WSAEWOULDBLOCK) {
	fprintf(stderr,"***select error: %d\n",error);
	close(socket);
	return -1;
      }
    }
#else
    if (select(FD_SETSIZE,NULL,&writefds,NULL,&seltimeout) < 0) {
      if (errno != EWOULDBLOCK) {
	perror("***");
	close(socket);
	return -1;
      }
    }
#endif

    else {

#ifdef borland
      if (FD_ISSET(socket,&exceptfds)) {
	close(socket);
	return -1;
      }
#endif
      
      if (FD_ISSET(socket ,&writefds)) {

	if ((nchar = (send(socket,(buffer+bytessend),
			   bufferlen-bytessend,0))) < 0) {
#ifdef borland
	  error = WSAGetLastError();
	  fprintf(stderr,"***send error: %d\n", error);
#else
	  perror("***");
#endif
	  close(socket);
	  return -1;
	} else {
	  bytessend += nchar;
	  if (bytessend == bufferlen)
	    break;
	}
      }
      
      /* check timeout */
      if (timeout) {
	time(&end);
	eltime = difftime(end,start);
      } 
      
    } /* else */
  } /* if select */
  
  if (bytessend == bufferlen) {
    return 1;
  } else {
    return 0;
  }
}


/*
 *   FUNCTION NAME
 *    hear
 *
 *   DESCRIPTION
 *    read a message or data from server.
 *
 *   ARGUMENTS
 *    msg - The message
 *    timeout - Timeout in s for TCP/IP
 *    blocking -  0: socket blocking
 *               >0: socket non blocking
 *    socket: TCP/IP socket
 *
 *   RETURN VALUE
 *    boolean error flag (0 = error; 1 = ok)
 *
 *   GLOBALS, ENVIRONMENT VARIABLES, FILE
 *
 *   REMARK:
 *     
 *
 */
 
int hear (int socket, double timeout, unsigned long blocking, char* msg)
{
  int             i,nchar,error;
  int             fsflags;
  char            recvbytes[RECORD_SIZE+1];
  struct timeval  seltimeout;
  double          sec;
  fd_set          readfds,exceptfds;
  unsigned long   mode;

 
  /* set socket options */
#ifdef borland
  if (blocking)
    mode = 0;
  else
    mode = 1;    
  if (ioctlsocket(socket,FIONBIO,&mode) < 0) {
    error = WSAGetLastError();
    fprintf(stderr,"***Error setting options %d\n",error);
  }

#else
#ifdef LINUX
  /* set socket options */
  if (fsflags = fcntl(socket,F_GETFL,0) < 0) {
    perror("***Error getting options***");
    return 0;
  };

  if (blocking)
    fsflags &= ~O_NONBLOCK;
  else
    fsflags |= O_NONBLOCK;
		
  if (fcntl(socket,F_SETFL,fsflags) < 0) {
    perror("***Error setting options***");
    return 0;
  }
#else
  if (blocking)
    mode = 0;
  else
    mode = 1;    
  if (ioctl(socket,FIONBIO,&mode) < 0) {
    perror("***");
  }
#endif
#endif
  
  /* initialize readfds */
  FD_ZERO(&readfds);
  FD_ZERO(&exceptfds);

  if (timeout) {
    seltimeout.tv_usec=(int) floor(modf(timeout,&sec))*1000000;
    seltimeout.tv_sec= (int) sec;
  }

  FD_SET(socket,&readfds);

#ifdef borland
  FD_SET(socket,&exceptfds);
  if (select(socket+1,&readfds,NULL,&exceptfds,&seltimeout) < 0) {
    error = WSAGetLastError();
    if (error != WSAEWOULDBLOCK) {
      fprintf(stderr,"***select error: %d\n", error);
      close(socket);
      return 0;
    }
  }
#else
  if (select(FD_SETSIZE,&readfds,NULL,NULL,&seltimeout) < 0) {
    if (errno != EWOULDBLOCK) {
      perror("select ***");
      close(socket);
      return 0;
    }
  }
#endif
   
  else {

#ifdef borland
    if (FD_ISSET(socket,&exceptfds)) {
      close(socket);
      return 0;
    }
#endif
    
    if (FD_ISSET(socket ,&readfds)) {
      
      /* fill recvbytes with \0 */
      memset(&recvbytes,'\0',sizeof(recvbytes));
      
      /* read from socket */ 
      nchar=recv(socket,(char *) &recvbytes,RECORD_SIZE,0);
      
#ifdef debug
      printf("*** number of bytes read from socket: %d\n", nchar);
#endif
      
      switch (nchar) 
	{
	case 0: 
	  fprintf(stderr,"*** Connection closed by remote! ***\n");
	  close(socket);
	  return 0;
	  
	case -1:
#ifdef borland
	  error = WSAGetLastError();
	  fprintf(stderr,"*** Connection Error: %d\n", error);
#else
	  perror("***");
#endif
	  close(socket);
	  return 0;
	  
	default:
	  strcpy(msg,recvbytes);
	  
#ifdef debug
	  printf("hear: %s\n", msg);
#endif 
	  
	  return 1;
	  
	} /* end switch */
    } /* end if FD_ISSET */
  } /* else */
  
  return 0; 
}



int main (int argc, char *argv[]) 
{


  int                    sock;    /* Base socket                   */
  int                    port;    /* Port to listen to             */
  struct sockaddr_in     sin;     /* Local address                 */
#ifdef borland
  int                    wVersionRequested;
  WSADATA                wsaData;
  int                    error;
#endif
  char                   *host;
  int                    ncycle,icycle; 
  char                   text[RECORD_SIZE+1];
  char                   *buf,*line,*p;

  int                    i,l,n;
  fd_set                 readfds;
  struct timeval         seltimeout;
  unsigned long          mode;
  int                    fd_stdin;

  int nchar;

  /* Process the input command line : accept three arguments: 
     hostname, portnumber, ncycle */

  
  switch (argc)
  {
    case 4:
      host = argv[1];
      if ((port = atoi (argv[2])) <= IPPORT_RESERVED) {
        fprintf(stderr,"Invalid port number\n");
        fprintf(stderr,"It must be a number greater than %d\n",
        IPPORT_RESERVED);
        exit (2);
      }
      ncycle = atoi(argv[3]);
      break;

    default:
      fprintf(stderr,"Wrong number of arguments\n");
      fprintf(stderr,"Usage: %s %s\n", argv[0]," [address,port,ncycle]");
      exit (2);
  }
  


#ifdef borland
  wVersionRequested = MAKEWORD(2, 2); 
  error = WSAStartup(wVersionRequested, &wsaData); 
 
  if (error != 0) { 
      /* Tell the user that we couldn't find a useable winsock.dll.     */ 
      fprintf(stderr,"could not find a useable winsock.dll\n");
      exit (2); 
  }
 
  /* Confirm that the Windows Sockets DLL supports 2.2.*/ 
  /* Note that if the DLL supports versions greater */ 
  /* than 2.2 in addition to 2.2, it will still return */ 
  /* 2.2 in wVersion since that is the version we requested. */ 
 
  if ( LOBYTE( wsaData.wVersion ) != 2 || HIBYTE( wsaData.wVersion ) != 2 ) { 
      /* Tell the user that we couldn't find a useable winsock.dll. */ 
      fprintf(stderr,"could not find a useable winsock.dll\n");
      WSACleanup(); 
      exit (2); 
  } 
  /* The Windows Sockets DLL is acceptable. Proceed. */ 
#endif


  /* allocate memory for buffer */
  buf = malloc(RECORD_SIZE+1); 
  strcpy(buf,"");

  sock = 0;
  if (!srvconn(host,port,1,&sock))
    exit(2);
    
#ifdef borland
  fd_stdin = fileno(stdin);
  if (filelength(fd_stdin) >= 0) {
    while (( nchar = (read(fileno(stdin),text,RECORD_SIZE))) > 0) {
    
      /* add \0 */
      text[nchar] = '\0';

      /* send text */
      if (!say(sock,1,0,text))
	goto end;
    }
  }

#elif VMS
  while (( nchar = (read(STDIN_FILENO,text,RECORD_SIZE))) > 0) {

    /* add \0 */
    text[nchar] = '\0';
    
    if (!say(sock,1,0,text))
      goto end;
  }
 
#else
  /* initialize readfds */
  FD_ZERO(&readfds);

  seltimeout.tv_usec= 0;
  seltimeout.tv_sec= 0;

  FD_SET(STDIN_FILENO,&readfds);

  if (select(FD_SETSIZE,&readfds,NULL,NULL,&seltimeout) < 0) {
    perror("***");
    goto end;
  }

  /* read text to send from standard input */
  if (FD_ISSET(STDIN_FILENO,&readfds)) {

    while (( nchar = (read(STDIN_FILENO,text,RECORD_SIZE))) > 0) {

      /* add \0 */
      text[nchar] = '\0';

      if (!say(sock,1,0,text))
	goto end;
    }
  }
#endif
  
  if (ncycle)
    icycle = 0;
  else
    icycle = -1;

  /* receive data */
  while (icycle < ncycle) {
    if (hear(sock,40,0,text) > 0) {

      /* reallocate memory for buffer */
      realloc(buf,strlen(buf) + strlen(text) + 1);

      /* copy text to buf */
      strcat(buf,text);

      /* test for end of cycle = "-CRLF" */
      if (strlen(buf) >= 3) {
	p = strrchr(buf,*LF);
	if (p == NULL)
	  continue;
	if ((*(p-2) == '-') && (*(p-1) == *CR) && (*p == *LF)) {
	  if (ncycle)
	    ++icycle;

	  /* remove CR and print lines to standard output */
	  line = strtok(buf, CR);
	  while (line != NULL) {
	    fprintf(stdout,"%s",line);
	    line = strtok(NULL, CR);
	  }

	  /* clear buffer */
	  strcpy(buf,"");
	}
      }
    } else {
      break;
    }
  }

  end:
     
  free(buf);
  close(sock);
  return 1;
}

